#include "mex.h"
#include "csources/utils.h"
#include "csources/DiskUtils.h"
#include "csources/Data.h"
#include "csources/MXData.h"

#include "csources/FeatureExtraction.h"

#include <cstring>
#include <cmath>

using namespace std;

void mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
	if((nlhs >= 1) && (nrhs  >= 11)) {

		const mxArray* mx_X	           = prhs[0];
		const mxArray* mx_image_width      = prhs[1];
		const mxArray* mx_image_height     = prhs[2];
		const mxArray* mx_region_width     = prhs[3];
		const mxArray* mx_region_height    = prhs[4];
		const mxArray* mx_tile_width       = prhs[5];
		const mxArray* mx_tile_height      = prhs[6];
		const mxArray* mx_padding_value    = prhs[7];
		const mxArray* mx_codebook         = prhs[8];
		const mxArray* mx_global_mean      = prhs[9];

		const mxArray* mx_mode             = prhs[10];

		//------ OMP case ------
		
		const mxArray* mx_num_components   = prhs[11];
		const mxArray* mx_schedule         = (nrhs > 12) ? prhs[12] : NULL;
		const mxArray* mx_n_desired        = (nrhs > 13) ? prhs[13] : NULL;
		const mxArray* mx_noise_level      = (nrhs > 14) ? prhs[14] : NULL;
		const mxArray* mx_use_dc_component = (nrhs > 15) ? prhs[15] : NULL;
		const mxArray* mx_output           = (nrhs > 16) ? prhs[16] : NULL;

		//------ Sparsent case -----

		const mxArray* mx_noise_var = mx_num_components;
		const mxArray* mx_beta      = mx_schedule;
		const mxArray* mx_sigma     = mx_n_desired;
		const mxArray* mx_tol       = mx_noise_level;
		const mxArray* mx_maxiter   = mx_use_dc_component;
		const mxArray* mx_num_flag  = mx_output;

		//------

		int image_width  = (int) *mxGetPr(mx_image_width);
		int image_height = (int) *mxGetPr(mx_image_height);

		int region_width  = (int) *mxGetPr(mx_region_width);
		int region_height = (int) *mxGetPr(mx_region_height);

		int tile_width  = (int) *mxGetPr(mx_tile_width);
		int tile_height = (int) *mxGetPr(mx_tile_height);

		int padding_value = (int) *mxGetPr(mx_padding_value);
	
		int num_components = (int) *mxGetPr(mx_num_components);

		double* global_mean = (mxGetM(mx_global_mean)*mxGetN(mx_global_mean) == tile_width*tile_height) ? double_array(tile_width*tile_height) : NULL;

		if (!global_mean) {
			PRINTF("size of global_mean does not equal tile_width*tile_height!!!\n");
			FLUSH;
			return;
		}

		memcpy(global_mean, mxGetPr(mx_global_mean), tile_width*tile_height*sizeof(double));

		int codebook_size  = mxGetN(mx_codebook);
		int codebook_rows  = mxGetM(mx_codebook);

		if (codebook_rows != tile_width*tile_height) {
			PRINTF("size of codebook elements does not equal tile_width*tile_height!!!\n");
			FLUSH;
			return;
		}

		int save_memory=0;
		DataDesc* X = MXDataDesc::Create(mx_X, save_memory);

		if (X->ssize() != image_width*image_height) {

			delete X;

			PRINTF("sample size does not equal image_width*image_height!!!\n");
			FLUSH;

			return;
		}

		double** codebook = m_create(codebook_size, codebook_rows);

		double** pcodebook = (double**) pointer_array(codebook_size);
		pcodebook[0] = mxGetPr(mx_codebook);
		for(int i=1; i < codebook_size; i++) pcodebook[i] = pcodebook[0] + i*codebook_rows;

		m_copy (codebook, pcodebook, codebook_size, codebook_rows);
		free_array(pcodebook);

		FeatureExtraction* fe_extr=NULL;

		int mode = (mx_mode != NULL) ? ((!mxIsEmpty(mx_mode)) ? (int)  *mxGetPr(mx_mode) : 0)    : 0;

		if (mode==0) {
			fe_extr = new SPFeatureExtraction (image_width, image_height, region_width, region_height,
						           tile_width, tile_height, padding_value,
			 			           codebook, codebook_size, global_mean);

			PRINTF("Using scalar product features ...\n");
			PRINTF("Image width: %i, Image height: %i, min/max region width: %i, min/max region height: %i, tile width: %i, tile height: %i, padding_value: %i\n",
				image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value);
			PRINTF("codebook size: %i\n", codebook_size);
			FLUSH;
		}
		else if (mode==1) {

			double noise_var = (mx_noise_var != NULL)  ? ((!mxIsEmpty(mx_noise_var)) ?  *mxGetPr(mx_noise_var) : 0.1)    : 0.1;
			double beta      = (mx_beta != NULL)       ? ((!mxIsEmpty(mx_beta))      ?  *mxGetPr(mx_beta)      : 1.0)    : 1.0;
			double sigma     = (mx_sigma != NULL)      ? ((!mxIsEmpty(mx_sigma))     ?  *mxGetPr(mx_sigma)     : 0.316)  : 0.316;

			double tol   = (mx_tol     != NULL) ? ((!mxIsEmpty(mx_tol))     ?  *mxGetPr(mx_tol)           : 0.01)  : 0.01;
			int maxiter  = (mx_maxiter != NULL) ? ((!mxIsEmpty(mx_maxiter)) ? (int)  *mxGetPr(mx_maxiter) : 100)    : 100;
			int num_flag = (mx_num_flag != NULL) ? ((!mxIsEmpty(mx_num_flag)) ? (int)  *mxGetPr(mx_num_flag) : 0)    : 0;

			fe_extr = new SparsenetFeatureExtraction (image_width, image_height, region_width, region_height,
				                                 tile_width, tile_height, padding_value,
								 codebook, codebook_size, 
				    				  noise_var, beta, sigma, tol, maxiter, num_flag, global_mean); 

			PRINTF("Performing Sparsenet feature extraction ...\n");
			PRINTF("Image width: %i, Image height: %i, min/max region width: %i, min/max region height: %i, tile width: %i, tile height: %i, padding_value: %i\n",
				image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value);
			PRINTF("codebook size: %i, noise_var: %g, beta: %g, sigma: %g, tol: %g, maxiter: %i, num_flag: %i\n", codebook_size, noise_var, beta, sigma, tol, maxiter, num_flag);
			FLUSH;

		}
		else if (mode==2) {
			int n_desired        = (int) (mx_n_desired != NULL)        ? (int) *mxGetPr(mx_n_desired)        : 1; 
			double noise_level   = (mx_noise_level != NULL)            ? *mxGetPr(mx_noise_level)            : 0; 
			int use_dc_component = (int) (mx_use_dc_component != NULL) ? (int) *mxGetPr(mx_use_dc_component) : 0; 
			int do_output        = (int) (mx_output != NULL)           ? (int) *mxGetPr(mx_output)           : 1;

			int* schedule     = NULL;
			int schedule_size = (mx_schedule != NULL) ? mxGetN(mx_schedule)*mxGetM(mx_schedule) : 0;

			if (schedule_size < num_components) schedule_size = 0;
		
			if (schedule_size) { 
				schedule = int_array(schedule_size);

				if (mxGetClassID (mx_schedule) == mxINT32_CLASS) memcpy(schedule, (int*) mxGetData(mx_schedule), schedule_size*4); 
				else double2int (schedule, mxGetPr(mx_schedule), schedule_size);
			}
			else {
				schedule = int_array(num_components); 

				iassign(schedule, 1, num_components);
			}

			PRINTF("Performing OMP feature extraction ...\n");
			PRINTF("Image width: %i, Image height: %i, min/max region width: %i, min/max region height: %i, tile width: %i, tile height: %i, padding_value: %i\n",
				image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value);
			PRINTF("codebook size: %i, number of non-zero components: %i\n", codebook_size, num_components);
			FLUSH;

			OMPFeatureExtraction* fe_extr = new OMPFeatureExtraction (image_width, image_height, region_width, region_height,
						   		         	  tile_width, tile_height, padding_value,
				  				         	  codebook, codebook_size, num_components,
										  noise_level, schedule, n_desired, global_mean);
			free_array(schedule);
		}

		DataDesc* features = fe_extr->Extract(X);
		int num_features = features->ssize();

		double** FE = NULL;
		features->GetAsMatrix(&FE);
		delete features;

		mxArray* mx_FE = mxCreateNumericMatrix (num_features, X->size(), mxDOUBLE_CLASS, mxREAL);
		m_to_vec(mxGetPr(mx_FE), FE, X->size(), num_features);

		plhs [0] = mx_FE;

		m_free(codebook);

		delete X;

		delete fe_extr;
		m_free(FE);

		free_array(global_mean);
	}
	else mexPrintf("Wrong number of input/output parameters (%i/%i). \n", nrhs, nlhs);
}

