#include "Data.h"
#include "FeatureExtraction.h"
#include "utils.h"
#include "csources/SPcgf.h"
#include "csources/ohl.h"

#include <stdlib.h>
#include <cmath>
#include <cstring>


RandomTileExtraction::RandomTileExtraction(int image_width, int image_height, int tile_width, int tile_height, int number_of_random_tiles) 
	: image_width(image_width), image_height(image_height), tile_width(tile_width), tile_height(tile_height), number_of_random_tiles(number_of_random_tiles) {
}

void RandomTileExtraction::reshape2d (double**** r_M, DataDesc* data) {

	int ssize = data->ssize();

	double*** M = (double***) pointer_array(data->size());
	*r_M = M;

	for (long i=0; i < data->size(); i++) {
		double* p = NULL;
		data->GetSample(&p, i);

		double** I = m_create(image_width, image_height);
		M[i] = I;

		for (long j=0; j < image_width; j++)  memcpy(I[j], p+j*image_height, image_height*sizeof(double));
	}
}


void RandomTileExtraction::GetAllTiles(double** dest, double** src) {

	int pos=0;
	for (int i=0; i <= image_width - tile_width; i++) {
		for (int j=0; j <= image_height - tile_height; j++) {
			for (int k=0;k < tile_width; k++) memcpy(dest[pos] + k*tile_height, src[i + k] + j, tile_height*sizeof(double));
			
			pos++;
		}
	}	
}

DataDesc* RandomTileExtraction::Extract(DataDesc* data) {

	int tile_size = tile_width*tile_height;

	StandardDataDesc* random_tiles = new StandardDataDesc(m_create(number_of_random_tiles, tile_size), tile_size, number_of_random_tiles);

	// convert the vector representation of the images into a two-dimensional representation
	double*** IMAGES = NULL;
	reshape2d(&IMAGES, data);

	// Matrix that contains each tile of the image as a vector
	double** TILES        = m_create((image_width-tile_width+1)*(image_height-tile_height+1), tile_size);

	// extract a number of random tiles out of each image
	int tiles_per_image = ceil(((double)number_of_random_tiles)/((double) data->size()));

	int pos       = 0;
 	int cur_image = 0;	
	while (pos < number_of_random_tiles) {
		int cur_pos = 0;

		GetAllTiles(TILES, IMAGES[cur_image]);

		while ((cur_pos < tiles_per_image) && (pos < number_of_random_tiles)) {
			double* cur_tile = NULL;
			random_tiles->GetSample(&cur_tile, pos);

			// select a random tile from the current image
			int select = floor(((double)image_width-tile_width+1)*((double)image_height-tile_height+1)*((double)rand())/((double) RAND_MAX + 1.0));
			// copy the selected tile to the data descriptor
			memcpy(cur_tile, TILES[select], tile_size*sizeof(double));

			cur_pos++;
			pos++;
		}
		cur_image++;
	}

	m_free(TILES);

	for (long i=0; i < data->size(); i++) m_free(IMAGES[i]);
	free_array(IMAGES); 

	return random_tiles;
}

void RandomTileExtraction::ComputeGlobalMean(double* global_mean, DataDesc* data) {

	memset(global_mean, 0, data->ssize()*sizeof(double));

	for (long i=0; i < data->size(); i++) {
		double* cur_sample = NULL;
		data->GetSample(&cur_sample, i);

		double local_mean = mean(cur_sample, data->ssize());
		ssub(cur_sample, cur_sample, local_mean, data->ssize());

		add(global_mean, global_mean, cur_sample, data->ssize());
	}

	sdivide(global_mean, global_mean, data->size(), data->ssize());

	for (long i=0; i < data->size(); i++) {
		double* cur_sample = NULL;
		data->GetSample(&cur_sample, i);

		sub(cur_sample, cur_sample, global_mean, data->ssize());
	}
}


RegionFeatureExtraction::RegionFeatureExtraction (int image_width, int image_height, int region_width, int region_height,
		      int tile_width, int tile_height, int padding_value) : image_width(image_width), image_height(image_height), region_width(region_width), region_height(region_height),
							 tile_width(tile_width), tile_height(tile_height), padding_value(padding_value) {

	vert_padding = ceil(tile_height/2);
	horz_padding = ceil(tile_width/2);

	width  = image_width  + 2*horz_padding;
	height = image_height + 2*vert_padding;

	int width_count  = image_width/region_width;
	int height_count = image_height/region_height;
	number_of_regions = width_count*height_count; 
}

void RegionFeatureExtraction::reshape2d (double**** r_M, DataDesc* data) {
		
	int ssize = data->ssize();

	double*** M = (double***) pointer_array(data->size());
	*r_M = M;

	for (long i=0; i < data->size(); i++) {
		double* p = NULL;
		data->GetSample(&p, i);

		double** I = m_create(width, height);
		M[i] = I;
		m_assign(I, width, height, padding_value);

		for (long j=0; j < image_width; j++)  memcpy(I[j + horz_padding] + vert_padding, p+j*image_height, image_height*sizeof(double));
	}
}

void  RegionFeatureExtraction::GetAllTiles(double** dest, double** src) {

	int pos=0;
	for (int i=0; i < image_width; i++) {
		for (int j=0; j < image_height; j++) {
			for (int k=0;k < tile_width; k++)  memcpy(dest[pos] + k*tile_height, src[i + k] + j, tile_height*sizeof(double));
			
			pos++;
		}
	}	
}

void RegionFeatureExtraction::GetRegionTiles(double** dest, double** src) {

	int temp_width  = (image_width/region_width)*region_width;
	int temp_height = (image_width/region_height)*region_height;

	int pos=0;
	for (int i=0; i < temp_width; i+= region_width) {
		for (int j=0; j < temp_height; j += region_height) {
			for (int k=0;k < region_width; k++) memcpy(dest[pos] + k*region_height, src[i + k] + j, region_height*sizeof(double));
			
			pos++;
		}
	}
}

DataDesc* RegionFeatureExtraction::Extract(DataDesc* data) {

	//PRINTF("number_of_regions: %i\n", number_of_regions);
	//PRINTF("region_width: %i, region_height: %i\n", region_width, region_height);

	StandardDataDesc* features = new StandardDataDesc(m_create(data->size(),number_of_regions*EncodingSize()*RegionEncodingSize()),
								number_of_regions*EncodingSize()*RegionEncodingSize(),data->size());

	// convert the vector representation of the images into a two-dimensional representation
	double*** IMAGES = NULL;
	reshape2d(&IMAGES, data);

	// Matrix that contains each tile of the image as a vector
	double** TILES        = m_create(image_width*image_height,tile_width*tile_height);
	// Matrix that contains each region of the coefficient plane as a vector
	double** REGION_TILES = m_create(number_of_regions, region_width*region_height); 
	m_zero (REGION_TILES, number_of_regions, region_width*region_height);

	// 3-dimensional matrix of coefficient planes, number of coefficient planes equals the size of the encoding
	double*** ENCODING_PLANES = (double***) pointer_array(EncodingSize());
	for (long i=0; i < EncodingSize(); i++) ENCODING_PLANES[i] = m_create(image_width, image_height); 

	double old_percent = 0.0;
	for (long i=0; i < data->size(); i++) {
		double percent = floor( ((double)i)*100/((double) data->size()));
		if (old_percent != percent) {
			old_percent = percent;
			PRINTF("%g %% \n", percent);
			FLUSH;
		}

		double* f = NULL;
		features->GetSample(&f, i);	

		// Perform tiling of  the current image
		GetAllTiles(TILES, IMAGES[i]);	

		// Perform encoding of the tiles (for example OMP)
		Encode(ENCODING_PLANES, TILES);	

		for (long k=0; k < EncodingSize(); k++) {
			// Extract regions of the coefficient planes (for example min/max)
			GetRegionTiles(REGION_TILES, ENCODING_PLANES[k]);

			EncodeRegions(f + k*number_of_regions*RegionEncodingSize(), REGION_TILES);
		}
	}

	m_free(TILES);
	m_free(REGION_TILES);

	for (long i=0; i < data->size(); i++) m_free(IMAGES[i]);
	free_array(IMAGES); 

	for (long i=0; i < EncodingSize(); i++) m_free(ENCODING_PLANES[i]);
	free_array(ENCODING_PLANES);

	return features;
}

int RegionFeatureExtraction::ImageWidth(void) { return image_width; }
int RegionFeatureExtraction::ImageHeight (void) { return image_height; }

int RegionFeatureExtraction::TileWidth(void) { return tile_width; }
int RegionFeatureExtraction::TileHeight(void) { return tile_height; }
int RegionFeatureExtraction::RegionSize(void) { return region_width*region_height; }
int RegionFeatureExtraction::NumberOfRegions(void) { return number_of_regions; }

//--------------------------------------------------------------------------------------------------------

OMPFeatureExtraction::OMPFeatureExtraction (int image_width, int image_height, int region_width, int region_height,
		         int tile_width, int tile_height, int padding_value,
			 double** codebook, int codebook_size, int num_components, double noise_level, int* sched, int n_desired,
			 double* global_mean) 
			: RegionFeatureExtraction (image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value), 
			  codebook(codebook), codebook_size(codebook_size), num_components(num_components), noise_level(noise_level), schedule(NULL), n_desired(n_desired), 
			  global_mean(global_mean) {

	schedule = int_array(num_components);	
	memcpy(schedule, sched, num_components*sizeof(int));
}

OMPFeatureExtraction::~OMPFeatureExtraction (void) {
	free_array(schedule);
}

// Returns the size of the encoding of each tile (number of encoding planes)
int OMPFeatureExtraction::EncodingSize(void) {
	return codebook_size;
}

// Performs the encoding of the tiles 
void OMPFeatureExtraction::Encode (double*** ENCODING_PLANES, double** TILES) {

	// remove local and global mean from data
	for (long i=0; i < ImageWidth()*ImageHeight(); i++) {
		double local_mean = mean(TILES[i], TileWidth()*TileHeight());
		ssub(TILES[i], TILES[i], local_mean, TileWidth()*TileHeight());
		sub(TILES[i], TILES[i], global_mean, TileWidth()*TileHeight());
	}

	int isOwner=0;	
	StandardDataDesc* data = new StandardDataDesc (TILES,TileWidth()*TileHeight(), ImageWidth()*ImageHeight(),isOwner);

	int use_dc_component = 0;
	int output = 0;

	double*** P = NULL;
	double*** R = NULL;

	bop(&P, &R, codebook, num_components, noise_level, data, codebook_size, schedule, n_desired, use_dc_component, output);

	for (long k=0; k < EncodingSize(); k++) {
		double** current_plane = ENCODING_PLANES[k];

		long pos=0;
		for(long i=0; i < ImageWidth(); i++) {
			for (long j=0; j < ImageHeight(); j++) {
				double** curP = P[pos];
				
				current_plane[i][j] = curP[0][k];

				pos = pos+1;
			}
		}
	}

	for (int i=0;i < data->num_samples();i++) m_free(P[i]);
        free_array(P);

	for (long i=0;i < data->num_samples();i++) m_free(R[i]);
        free_array(R);
	
	delete data;
}

int OMPFeatureExtraction::RegionEncodingSize(void) {
	return 2;
}

void OMPFeatureExtraction::EncodeRegions(double* dest, double** REGION_TILES) {

	for (long k=0; k < NumberOfRegions(); k++) {
		double* cur_region = REGION_TILES[k];
		dest[2*k]   = cur_region[dmin(cur_region, RegionSize())];
		dest[2*k+1] = cur_region[dmax(cur_region, RegionSize())];
	}
}

//--------------------------------------------------------------------------------------------------------

SPFeatureExtraction::SPFeatureExtraction (int image_width, int image_height, int region_width, int region_height,
		         		  int tile_width, int tile_height, int padding_value,
			 		  double** codebook, int codebook_size, double* global_mean) 
			: RegionFeatureExtraction (image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value), 
			  			   codebook(codebook), codebook_size(codebook_size), global_mean(global_mean) {
}

// Returns the size of the encoding of each tile (number of encoding planes)
int SPFeatureExtraction::EncodingSize(void) {
	return codebook_size;
}

// Performs the encoding of the tiles 
void SPFeatureExtraction::Encode (double*** ENCODING_PLANES, double** TILES) {

	// remove local and global mean from data
	for (long i=0; i < ImageWidth()*ImageHeight(); i++) {
		double local_mean = mean(TILES[i], TileWidth()*TileHeight());
		ssub(TILES[i], TILES[i], local_mean, TileWidth()*TileHeight());
		sub(TILES[i], TILES[i], global_mean, TileWidth()*TileHeight());
	}

	for (long k=0; k < EncodingSize(); k++) {
		double** current_plane = ENCODING_PLANES[k];

		long pos=0;
		for(long i=0; i < ImageWidth(); i++) {
			for (long j=0; j < ImageHeight(); j++) {
				// use scalar product between basis vector and tile
				current_plane[i][j] = scp(TILES[pos], codebook[k], TileWidth()*TileHeight() );

				pos = pos+1;
			}
		}
	}
}

int SPFeatureExtraction::RegionEncodingSize(void) {
	return 2;
}

void SPFeatureExtraction::EncodeRegions(double* dest, double** REGION_TILES) {

	for (long k=0; k < NumberOfRegions(); k++) {
		double* cur_region = REGION_TILES[k];
		dest[2*k]   = cur_region[dmin(cur_region, RegionSize())];
		dest[2*k+1] = cur_region[dmax(cur_region, RegionSize())];
	}
}

//--------------------------------------------------------------------------------------------------------

SparsenetFeatureExtraction::SparsenetFeatureExtraction (int image_width, int image_height, int region_width, int region_height,
		                                        int tile_width, int tile_height, int padding_value,
			 				double** codebook, int codebook_size, double noise_var, double beta, double sigma, double tol, int maxiter, int num_flag, double* global_mean) 
						: RegionFeatureExtraction (image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value), 
							  codebook(codebook), codebook_size(codebook_size), noise_var(noise_var), beta(beta), sigma(sigma), tol(tol), maxiter(maxiter), num_flag(num_flag),
							  global_mean(global_mean) {
}

// Returns the size of the encoding of each tile (number of encoding planes)
int SparsenetFeatureExtraction::EncodingSize(void) {
	return codebook_size;
}

// Performs the encoding of the tiles 
void SparsenetFeatureExtraction::Encode (double*** ENCODING_PLANES, double** TILES) {

	// remove local and global mean from data
	for (long i=0; i < ImageWidth()*ImageHeight(); i++) {
		double local_mean = mean(TILES[i], TileWidth()*TileHeight());
		ssub(TILES[i], TILES[i], local_mean, TileWidth()*TileHeight());
		sub(TILES[i], TILES[i], global_mean, TileWidth()*TileHeight());
	}

	int isOwner=0;	
	StandardDataDesc* data = new StandardDataDesc (TILES,TileWidth()*TileHeight(), ImageWidth()*ImageHeight(),isOwner);

	int use_dc_component = 0;
	int output = 0;

	double** P = m_create(data->num_samples(), codebook_size);
	//double*** R = NULL;

	//bop(&P, &R, codebook, num_components, noise_level, data, codebook_size, schedule, n_desired, use_dc_component, output);

	double lambda=1.0/noise_var;
	int outflag=0;
	SPcgf* my_cgf = new SPcgf(codebook, data->ssize(), codebook_size, lambda, beta, sigma, outflag);

	my_cgf->cgf(P, data, tol, maxiter, num_flag);

	for (long k=0; k < EncodingSize(); k++) {
		double** current_plane = ENCODING_PLANES[k];

		long pos=0;
		for(long i=0; i < ImageWidth(); i++) {
			for (long j=0; j < ImageHeight(); j++) {
				double* curP = P[pos];
				
				current_plane[i][j] = curP[k];

				pos = pos+1;
			}
		}
	}

	//for (int i=0;i < data->num_samples();i++) m_free(P[i]);
        //free_array(P);

	m_free(P);

	//for (long i=0;i < data->num_samples();i++) m_free(R[i]);
        //free_array(R);
	
	delete data;
}

int SparsenetFeatureExtraction::RegionEncodingSize(void) {
	return 2;
}

void SparsenetFeatureExtraction::EncodeRegions(double* dest, double** REGION_TILES) {

	for (long k=0; k < NumberOfRegions(); k++) {
		double* cur_region = REGION_TILES[k];
		dest[2*k]   = cur_region[dmin(cur_region, RegionSize())];
		dest[2*k+1] = cur_region[dmax(cur_region, RegionSize())];
	}
}

