class DataDesc;

// abtract basis class for all feature extraction classes
class FeatureExtraction {
public:
	virtual DataDesc* Extract(DataDesc* data) = 0;
};

// this class provides random tiles of given images. These radom tiles are used in order to learn a basis for example using the sparsenet algorithm 
// or SCNG (sparse coding neural gas)
class RandomTileExtraction : public FeatureExtraction {
public:
	RandomTileExtraction(int image_width, int image_height, int tile_width, int tile_height, int number_of_random_tiles);

	void reshape2d (double**** r_M, DataDesc* data);
	void GetAllTiles(double** dest, double** src);
	virtual DataDesc* Extract(DataDesc* data);

	// global_mean is a double array of size data->ssize()
	void ComputeGlobalMean(double* global_mean, DataDesc* data);

private:
	int image_width;
	int image_height;
	int tile_width;
	int tile_height;
	int number_of_random_tiles;
};

class RegionFeatureExtraction : public FeatureExtraction {
public:
	// image_width, image_height : width and height of images to be processed
	// region_width, region_height : size of the  regions
	// tile_width, tile_height : size of the feature extractors
	//
	RegionFeatureExtraction (int image_with, int image_height, int region_width, int region_height,
			      int tile_width, int tile_height, int padding_value);

	void reshape2d (double**** r_M, DataDesc* data);

	void GetAllTiles(double** dest, double** src);
	void GetRegionTiles(double** dest, double** src);
	virtual DataDesc* Extract(DataDesc* data);

	// Returns the size of the encoding of each tile (number of encoding planes)
	virtual int  EncodingSize(void) = 0;

	// Performs the encoding of the tiles 
	virtual void Encode (double*** ENCODING_PLANES, double** TILES) = 0;

	virtual int RegionEncodingSize(void) = 0;
	virtual void EncodeRegions(double* dest, double** REGION_TILES) = 0;

	int ImageWidth(void);
	int ImageHeight (void);

	int TileWidth(void);
	int TileHeight(void);
	int RegionSize(void);
	int NumberOfRegions(void);

private:
	int width;
	int height;

	int image_width;
	int image_height;

	int tile_width;
	int tile_height;

	int region_width;
	int region_height;
	int number_of_regions;

	int vert_padding;
	int horz_padding;
	double padding_value;
};

class SPFeatureExtraction : public RegionFeatureExtraction {
public:
	SPFeatureExtraction (int image_width, int image_height, int region_width, int region_height,
			      int tile_width, int tile_height, int padding_value,
			      double** codebook, int codebook_size, double* global_mean);

	// Returns the size of the encoding of each tile (number of encoding planes)
	virtual int  EncodingSize(void);

	// Performs the encoding of the tiles 
	virtual void Encode (double*** ENCODING_PLANES, double** TILES);

	virtual int RegionEncodingSize(void);

	virtual void EncodeRegions(double* dest, double** REGION_TILES);

private:
	double** codebook;
	int codebook_size;

	double* global_mean;
};

class OMPFeatureExtraction : public RegionFeatureExtraction {
public:
	OMPFeatureExtraction (int image_width, int image_height, int region_width, int region_height,
			      int tile_width, int tile_height, int padding_value,
			      double** codebook, int codebook_size, int num_components, double noise_level, int* schedule, int n_desired,
			      double* global_mean);

	~OMPFeatureExtraction(void); 

	// Returns the size of the encoding of each tile (number of encoding planes)
	virtual int  EncodingSize(void);

	// Performs the encoding of the tiles 
	virtual void Encode (double*** ENCODING_PLANES, double** TILES);

	virtual int RegionEncodingSize(void);

	virtual void EncodeRegions(double* dest, double** REGION_TILES);

private:
	double** codebook;
	int codebook_size;

	int num_components;
	double noise_level;
	int* schedule;
	int n_desired;
	double* global_mean;
};

class SparsenetFeatureExtraction : public RegionFeatureExtraction {
public:

	SparsenetFeatureExtraction (int image_width, int image_height, int region_width, int region_height,
	                            int tile_width, int tile_height, int padding_value,
				    double** codebook, int codebook_size, 
				    double noise_var, double beta, double sigma, double tol, int maxiter, int num_flag, double* global_mean); 


	// Returns the size of the encoding of each tile (number of encoding planes)
	virtual int  EncodingSize(void);

	// Performs the encoding of the tiles 
	virtual void Encode (double*** ENCODING_PLANES, double** TILES);

	virtual int RegionEncodingSize(void);

	virtual void EncodeRegions(double* dest, double** REGION_TILES);

private:
	double** codebook;
	int codebook_size;

	double* global_mean;

	double noise_var;
	double beta;
	double sigma;
	double tol;
	int maxiter;
	int num_flag;
};
