/*

This is a slightly modified version, i.e., a "C++ - ified" version, of cgf.c taken from the Sparsenet package
by Bruno Olshausen:
https://redwood.berkeley.edu/bruno/sparsenet. 

See also

Olshausen BA, Field DJ (1997). Sparse coding with an overcomplete
basis set: A strategy employed by V1?  Vision Research, 37, 3311-3325.
ftp://redwood.ucdavis.edu/pub/papers/VR.pdf
 
Olshausen BA, Field DJ (1996). Emergence of simple-cell receptive
field properties by learning a sparse code for natural images.
Nature, 381, 607-609.
ftp://redwood.ucdavis.edu/pub/papers/nature-paper.pdf

cgf.c:	conj. grad. routine for finding optimal s - fast!

*/
#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <cstring>

#include "csources/SPcgf.h"
#include "csources/utils.h"
#include "csources/Data.h"

#define sgn(x) (x>0 ? 1 : (x<0 ? -1 : 0))

/* Define indexing macros for matricies */

/* L = dimension of input vectors
 * M = number of basis functions
 */

#define A_(i,j)		A[(i) + (j)*L]		// A is L x M
#define X_(i,n)		X[(i) + (n)*L]		// X is L x npats

#define Sout_(i,n)	Sout[(i) + (n)*M]	// S is M x npats
#define Sin_(i,n)	Sin[(i) + (n)*M]	// S is M x npats

#define AtA_(i,j)	AtA[(i) + (j)*M]	// AtA is M x M

using namespace std;

SPcgf::SPcgf(double** codebook, int ssize, int codebook_size, 
             double lambda, double beta, double sigma, int outflag) : L(ssize), M(codebook_size), lambda(lambda), beta(beta), sigma(sigma), itmax(100), outflag(outflag), codebook(codebook), 
								      k1(0), k2(0), k3(0), gcount(0), fcount(0) {

	A = double_array(M*L);

	for (int i=0; i < M; i++) memcpy(A + i*L, codebook[i], L*sizeof(double));

	x      = double_array(L);
  	s0     = double_array(M);
  	d      = double_array(M);
  	AtA    = double_array(M*M);
  	Atx    = double_array(M);

  	// Calc  A'*A
  	for (int i = 0; i < M; i++) {
    		double* Ai=A+i*L;
    		for (int j = 0; j < M; j++) {
      			double* Aj=A+j*L;
      			double sum=0.0;
      			for (int k = 0; k < L; k++) sum += Ai[k]*Aj[k];
      			AtA_(i,j) = sum;
    		}
  	}
}

SPcgf::~SPcgf(void) {

	free_array(x);
	free_array(s0);
	free_array(d);
	free_array(AtA);
	free_array(Atx);

	free_array(A);
}

float SPcgf::init_f1dim(float* s1, float* d1) {
  	register int		i,j;
  	register double	As,Ag,sum;
  	register float	fval;
  
  	for (i=0; i < M; i++) {
    		s0[i]=s1[i+1];
    		d[i]=d1[i+1];
  	}

  	k1=k2=k3=0;
  	for (i=0; i<L; i++) {
    		As=Ag=0;
    		for (j=0; j<M; j++) {
      			As += A_(i,j) * s0[j];
      			Ag += A_(i,j) * d[j];
    		}
    		k1 += As*(As - 2*x[i]);
    		k2 += Ag*(As - x[i]);
    		k3 += Ag*Ag;
  	}
  	k1 *= 0.5*lambda;
  	k2 *= lambda;
  	k3 *= 0.5*lambda;

  	fval = k1;

  	sum=0;
  	for (i=0; i<M; i++) sum += sparse(s0[i]/sigma);

  	fval += beta*sum;

  	fcount++;

  	return(fval);
}

float SPcgf::f1dim(float alpha) {
  	double 	sum;
  	float		fval;

  	fval = k1+(k2+k3*alpha)*alpha;

  	sum=0;
  	for (int i=0; i<M; i++) sum += sparse((s0[i]+alpha*d[i])/sigma);
  	
  	fval += beta*sum;

  	fcount++;

  	return(fval);
}


// Gradient evaluation used by conj grad descent
void SPcgf::dfunc(float* p,float* grad) {
	register int		i,j;
  	register double	sum,*cptr,bos=beta/sigma;
  	register float	*p1;

  	p1=&p[1];

  	for (i=0; i<M; i++) {
    		cptr=AtA+i*M;

    		sum=0;
    		for (j=0; j<M; j++)  sum += p1[j] * *cptr++;
    		
    		grad[i+1] = lambda*(sum - Atx[i]) + bos*sparse_prime((double)p1[i]/sigma);
  	}
  	gcount++;
}

double SPcgf::sparse(double x) {
	return(log(1.0+x*x));
}

double SPcgf::sparse_prime(double x) {
	return(2*x/(1.0+x*x));
}

#include "nrf/nrutil.h"

void SPcgf::cgf(double** P, DataDesc* X, double tol, int maxiter, int numflag) {
	double* x = double_array(X->ssize());
	double* p_init = double_array(M);

	double nits=0;
	double nf=0;
	double ng=0;

	itmax = maxiter;

	for (int i=0; i < X->size(); i++) {

		int npats=1;

		int* NZ=NULL;
		X->CopySample(x, &NZ, i);

		// compute initial guess for coefficients
		for (int j=0;j < M; j++) p_init[j] = scp(x, codebook[j], X->ssize())/scp(codebook[j],codebook[j], X->ssize());

		double* Sout = P[i];
		double* Sin  = p_init;

		do_cgf(Sout, &nits, &nf, &ng, Sin, x, npats, tol, maxiter, numflag); 
	}

	free_array(x);
	free_array(p_init);
}

void frprmn(float p[],int n,float ftol,int* iter,float* fret,CGFTargetFunc* tfunc);

void SPcgf::do_cgf(double *Sout, double *nits, double *nf, double *ng, 
		double *Sin, double *X, int npats, 
		double tol, int maxiter, int numflag) {

  	double	sum;
  	float 	fret;
  	int 	niter;
	float 	*p;

  	p=vector(1,M);

  	for (int n = 0; n < npats; n++) {
    		if (numflag) {
      			fprintf(stderr,"\r%d",n+1);
      			fflush(stderr);
    		}

    		for (int l = 0; l < L; l++) x[l] = X_(l,n);

    		for (int m = 0; m < M; m++) {

      			// precompute Atx for this pattern
      			sum = 0.0;
      			for (int l = 0; l < L; l++)  sum += A_(l,m) * x[l];
      
      			Atx[m] = sum;

      			// copy initial guess
      			p[m+1] = Sin_(m,n);
    		}

    		fcount=gcount=0;

    		frprmn(p, M, (float)tol, &niter, &fret, this);

    		*nits += (double)niter;
    		*nf += (double)fcount;
    		*ng += (double)gcount;

    		if (outflag) {
      			fprintf(stdout,"\nfret=%f  niters=%d  fcount=%d  gcount=%d\n", fret, niter, fcount, gcount);
      			fflush(stdout);
    		}

  		// copy back solution
    		for (int m = 0; m < M; m++)  Sout_(m,n) = p[m+1];
	}

	free_vector(p,1,M);
}

#undef A_
#undef X_
#undef Sout_
#undef Sin_
#undef AtA_
