%
%function digit_recognition(nvalidation, ntrain, ntest, fe_mode, precompute_kernel)
%
% This experiment is described in [1]. We use the SoftDoubleMaxMinOver SVM
% training method [2]. An implementation of this method (CVMaxMinOver,
% MaxMinOverClassify) can be obtained from:
%    http://www.inb.uni-luebeck.de/tools-demos/softdoublemaxminover/DMMO.tar.gz
%
% nvalidation : number of training images to be used during validation
% ntrain      : number of training images to be used for final training
% ntest       : number of test images to be used
%
% In order to reproduce the results described in [1], set
% nvalidation=10000, ntrain=60000 and ntest=10000.
%
% Warnings:
%
% (I) For fe_mode=1 (i.e., determine basis coefficients by minimizing
% the Sparsenet target function via gradient descent) this script is
% computationally very demanding, i.e., for all training (60000) and test
% (10000) samples it may take several days to run this script. There
% are three simple approaches to reduce the computational demand or computation time:
%    1. Use fewer data (for example nvalidation=ntrain=ntest=1000)
%    2. Distribute the computation over many computers (this was done for [1])
%    3. Use simple overlap in order to determine the coefficients of the
%       basis functions (set fe_mode to 0, this also works quite good)
%
% (II) If many validation/training samples are to be used, it might be neccessary to
% set precompute_kernel to 0 (this depends on the memory configuration of
% your computer).
%
% [1] Kai Labusch, Erhardt Barth, and Thomas Martinetz. Simple Method for
%     High-Performance Digit Recognition Based on Sparse Coding. 
%     IEEE Transactions on Neural Networks, 19(11):1985-1989, 2008.
% [2] Thomas Martinetz, Kai Labusch, and Daniel Schneega??. 
%     SoftDoubleMaxMinOver: Perceptron-like Training of Support Vector 
%     Machines. IEEE Transactions on Neural Networks, 20(7):1061-1072, 2009.
%
function digit_recognition(nvalidation, ntrain, ntest, fe_mode, precompute_kernel)

%-----------

if ~exist('ntrain', 'var') || isempty(ntrain)
    nvalidation=1000;
end

if ~exist('ntrain', 'var') || isempty(ntrain)
    ntrain=1000;
end

if ~exist('ntest','var') || isempty(ntest)
    ntest=1000;
end

if ~exist('fe_mode', 'var') || isempty(fe_mode)
    fe_mode=1;
end

if ntrain < nvalidation
    ntrain = nvalidation;
end

if ~exist('precompute_kernel', 'var') || isempty(precompute_kernel)
    precompute_kernel=1; % precompute kernel matrix (do this only if enough memory is available!!!)
end

%-----------

load('mnist.mat');
filename = sprintf('spn_features_test_nvalidation(%i)_ntrain(%i)_ntest(%i)_fe(%i).mat', nvalidation, ntrain, ntest, fe_mode);

% randomly select ntrain training samples
P=randperm(ntrain);
X_train = double(X_train(:,P(1:ntrain))); %#ok<NODEF>
c_train = double(c_train(P(1:ntrain))); %#ok<NODEF>

%randomly select ntrain test samples
P=randperm(ntest);
X_test = double(X_test(:,P(1:ntest))); %#ok<NODEF>
c_test = double(c_test(P(1:ntest))); %#ok<NODEF>

image_width  = 28; % size of MNIST images is 28x28
image_height = 28;

num_basis = 169;   % number of
tile_width  = 13;  % size of basis functions (13x13)
tile_height = 13;

region_width  = 9;   % size of min/max regions (9x9)
region_height = 9;
padding_value = 255; % use this color in order to enlarge the images

% obtain an estimation of global_mean and scaling factor
number_of_random_tiles = 10000; 
[T, global_mean] = RFE(X_train, image_width, image_height, tile_width, tile_height, number_of_random_tiles);
scale_factor = sqrt(0.1)/sqrt(mean(var(T,0,1)));
clear T; % we do not need T anymore

% Parameters of the sparsenet algorithm
num_trials = 500; 
batch_size = 100;
noise_var  = 0.05;
beta       = 1.0;    %2.2;
sigma      = 0.316;
tol        = 0.01;
maxiter    = 10;
num_flag   = 0; 

% use sparsenet algorithm in order to learn the basis
C = sparsenet (scale_factor*X_train, image_width, tile_width, num_basis, num_trials, batch_size, noise_var, beta, sigma, scale_factor*global_mean');

save(filename); 

% extract training features
F_train = FE (X_train, image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value, ...
              C, global_mean, fe_mode, noise_var, beta, sigma, tol, maxiter, num_flag);          

save(filename);         

% extract test features
F_test = FE (X_test, image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value, ...
             C, global_mean, fe_mode, noise_var, beta, sigma, tol, maxiter, num_flag);

save(filename);         
          
% perform cross-validation in order to determine the best
% SVM-hyper-parameters
k               = 10; % 10-fold cross validation
kernel_type     = 1;  % use gaussian kernel
grid_count      = 5;  %
max_level       = 2;  %
KERNEL_CACHE_MB = 1000;
[alpha, b, SVs, cSV, ucl, sv_indizes, nf, k_params, bmerr, bm_svcount, C_hardness] = CVMaxMinOver (F_train(:,1:nvalidation), c_train(1:nvalidation), k, kernel_type,...
                                                                    grid_count, precompute_kernel, max_level, KERNEL_CACHE_MB);         
        
save(filename);

if ntrain > nvalidation % Perform final training using all training samples
    lparams=lparam_set(1,'kernel_type', kernel_type, 'KERNEL_CACHE_MB', KERNEL_CACHE_MB, 'C', C_hardness);
    [alpha, b, SVs, cSV, ucl, sv_indizes] = TrainMaxMinOver (F_train, c_train, lparam_get('learn_params', lparams,1), k_params, nf);
end

save(filename);

% perform final classification of test data
[decision, labels] = MaxMinOverClassify (alpha, b, SVs, ucl, kernel_type, k_params, F_test, nf);

save(filename);

errors = find(labels ~= c_test);

error_rate = length(errors)/length(c_test);
fprintf('Error rate: %g\n', error_rate);

save(filename);

return;