%
% This is a slightly modified version of the sparsenet.m file of the Sparsenet 
% package by Bruno Olshausen.
% See https://redwood.berkeley.edu/bruno/sparsenet/
%
%References:
% 
% Olshausen BA, Field DJ (1997). Sparse coding with an overcomplete
% basis set: A strategy employed by V1?  Vision Research, 37, 3311-3325.
% ftp://redwood.ucdavis.edu/pub/papers/VR.pdf
% 
% Olshausen BA, Field DJ (1996). Emergence of simple-cell receptive
% field properties by learning a sparse code for natural images.
% Nature, 381, 607-609.
% ftp://redwood.ucdavis.edu/pub/papers/nature-paper.pdf
%
%
% A = sparsenet (IMAGES, image_size, tile_size, num_basis, num_trials,...
%                batch_size, noise_var, beta, sigma, global_mean)
%
% IMAGES      : images to be processed
% image_size  : size of images 
% tile_size   : size of random tiles 
% num_basis   : number of basis vectors
% num_trials  : number of outer loop iterations
% batch_size  : number of images considered per loop iteration
% noise_var   : variance of the noise (|x-As|^2) (optional)
% beta        : steepness term for prior
% sigma       : scaling term for prior
% global_mean : globel mean of random patches where the local mean 
%               has been removed
function A = sparsenet (IMAGES, image_size, tile_size, num_basis, num_trials,...
                        batch_size, noise_var, beta, sigma, global_mean)

if ~exist('noise_var', 'var'),
    noise_var = 0.01;
end

if ~exist('batch_size','var'),
    batch_size = 500;
end

if ~exist('num_trials','var'),
    num_trials = 10000;
end

if ~exist('beta', 'var'),
    beta= 2.2;
end

if ~exist('sigma','var'),
    sigma=0.316;
end

if ~exist('save_file','var'),
    save_file = [];
end

eta      = 1.0;
tol      = .01;
VAR_GOAL = 0.1;
var_eta  = .001;
alpha    = .02;
maxiter  = 10;

% main loop
A = [];
old_percent=0;

fprintf('Performing Sparsenet training ...\n');

for t=1:num_trials
    
    new_percent = floor(t*100/num_trials);
    if new_percent ~= old_percent,
        fprintf('%i%%\n', new_percent);
        old_percent=new_percent;
    end
    
    X=RandomTiles(IMAGES, image_size, tile_size, batch_size);
        
    %remove local mean
    X = X - repmat( mean(X,1), size(X,1), 1);
    %remove global mean
    X = X - repmat(global_mean, 1, size(X,2));
    
    if isempty(A),
        A = rand(size(X,1), num_basis) - 0.5;
        A = A*diag(1./sqrt(sum(A.*A)));

	    [L M]=size(A);
    	gain=sqrt(sum(A.*A))';
    	S_var=VAR_GOAL*ones(M,1);
    end
    
    % calculate coefficients for these data via conjugate gradient routine    
    S=SPCGF(X, A, noise_var, beta, sigma, maxiter, tol);
    
    % calculate residual error
    E=X-A*S;
    
    % update bases
    dA=zeros(L,M);
    for i=1:batch_size
        dA = dA + E(:,i)*S(:,i)';
    end
    dA = dA/batch_size;
    
    A = A + eta*dA;
    
    % normalize bases to match desired output variance
    for i=1:batch_size
        S_var = (1-var_eta)*S_var + var_eta*S(:,i).*S(:,i);
    end
    gain = gain .* ((S_var/VAR_GOAL).^alpha);
    normA=sqrt(sum(A.*A));
    for i=1:M
        A(:,i)=gain(i)*A(:,i)/normA(i);
    end
    
    display_network(A, tile_size, tile_size, 1, 0);colormap(gray);
end
